## For a given window size, compute the moving average, and then compute by-row difference of a tabular data set
ma_diff <- function(data,ma_window_size){
  ## Compute moving average too smoothen noise from sensors a bit
  mav <- function(x, n = ma_window_size){as.numeric(stats::filter(x, rep(1 / n, n), sides = 2))}
  data_ma <- as.data.frame(apply(data,2,mav))
  
	## Row wise differencing of the resulting data
  data_diff <- rbind(NA,apply(data_ma,2,diff))

  data_diff
}

## Reshape tabular data into sequence blocks (arrays) to prepare for use in NN model
reshape_sequences_lag_event <- function(data,labels){
  samps <- nrow(data) #number of samples overall
  nfeats <- ncol(data) #number of features
  
  ## number of timesteps within which a detection of an event is still counted 
  ## as a positive detection
  timesteps_add <- 63
  lag_pattern <- 1:63
  
  x_seq <- array(dim=c(samps-(max(lag_pattern)+1)+1,length(lag_pattern)+1,nfeats))
  x_seq_labels_exclude <- logical(dim(x_seq)[1])
  x_seq_labels <- logical(dim(x_seq)[1])
  for(i in (max(lag_pattern)+1):(samps)){
    j <- i-max(lag_pattern)
    sel <- c(i,i-lag_pattern) 
    x_seq[j,,] <- data[sel,]
    x_seq_labels_exclude[j] <- any(labels[sel])
    ndx <- (i-timesteps_add):i
    ndx <- pmin(samps,pmax(1,ndx)) # limit to valid range
    x_seq_labels[j] <- any(labels[ndx])
  }
  list(x_seq=x_seq,x_seq_labels_exclude=x_seq_labels_exclude,x_seq_labels=x_seq_labels)
}

## Calculate F1 score
calculateScore <- function(observations, predictions){
  TruePositive <- sum(observations & predictions)
  FalsePositive <- sum(!observations & predictions)
  TrueNegative <- sum(!observations & !predictions)
  FalseNegative <- sum(observations & !predictions)
  
  PositivePredictiveValue <- TruePositive / sum(predictions)
  TruePositiveRate <- TruePositive / sum(observations)
  
  F1score <- (2 * PositivePredictiveValue * TruePositiveRate) / (PositivePredictiveValue + TruePositiveRate)
  
  if (is.nan(F1score)) F1score <- 0
  
  return(list(TP=TruePositive, FP=FalsePositive, TN=TrueNegative, FN=FalseNegative, SCORE= F1score))
}

## Optimize threshold for events (i.e., select the threshold which maximizes F1)
optimizeThreshold <- function(labels,errors){
  fopt <- function(tt){
    res <- calculateScore(labels,errors>tt)$SCORE
  }
  thresholds <- seq(from=min(errors),to=max(errors),length.out=1000)
  scores <- sapply(thresholds,fopt)
  bounds <- quantile(errors,probs=c(10^-4,1-10^-4))
  opt <- optimize(fopt,interval=bounds,maximum=TRUE)
  return(opt$maximum)
}
