runIF_test <- function(
  sample_size=486,	
  ntrees = 905,
  ndim=5,
  ntry=7,
  max_depth=2,
  ncols_per_tree=23,
  seed=1L
){
  set.seed(seed)
  
  source("functions.R")
  
  ## read data files
  data <- readRDS(file = "waterDataTraining.RDS")
  datatest <- readRDS(file = "waterDataTestingUpload.RDS")
  data <- rbind(data,datatest)
  datatest <- NULL
  gc()
  
  
  ## separate columns not directly passed to model
  data_label <- data$EVENT
  data_time <- data$Time
  data$Time <- NULL
  data$EVENT <- NULL
  
  ## differencing and moving average
  data <- ma_diff(data,5)
  
  ## replace NA values by moving average
  data <- na_locf(data)
  
  ## data split into training and validation
  train_index <- 1:round(nrow(data)*0.5)
  validation_index <- (round(nrow(data)*0.5)+1-64):nrow(data) 
  train_data <- data[train_index,]
  validation_data <- data[validation_index,]
  train_data_labels <- data_label[train_index]
  validation_data_labels <- data_label[validation_index]
  
  
  ## mean and standard deviation for scaling
  scale_mean <- apply(train_data, 2, mean,na.rm=T)
  scale_sd <- apply(train_data, 2, sd,na.rm=T)
  
  ## scale data
  train_data_scaled <- scale(train_data, center = scale_mean, scale = scale_sd)
  validation_data_scaled <- scale(validation_data, center = scale_mean, scale = scale_sd)
  
  ## number of features in the data
  nfeats <- ncol(train_data_scaled) #number of features
  
  ## compute sequence blocks
  result <- reshape_sequences_lag_event(train_data_scaled,train_data_labels)
  x_seq_train <- result$x_seq
  x_seq_train_labels <- result$x_seq_labels
  x_seq_train_labels_exclude <- result$x_seq_labels_exclude
  result <- NULL
  
  ## subsample to vary training data set
  nseq <- dim(x_seq_train)[1]
  subsamplel <- sample(c(TRUE, FALSE),nseq,prob = c(0.66,0.33),replace=TRUE)  
  x_seq_train <- x_seq_train[subsamplel,,]
  x_seq_train_labels <- x_seq_train_labels[subsamplel]
  x_seq_train_labels_exclude <- x_seq_train_labels_exclude[subsamplel]
  
  ## exclude from training data: sequences that contain events 
  x_seq_train_noevent <- x_seq_train[!x_seq_train_labels_exclude,,]
  
  ## generate validation sequence blocks
  result <- reshape_sequences_lag_event(validation_data_scaled,validation_data_labels)
  x_seq_validation <- result$x_seq
  x_seq_validation_labels <- result$x_seq_labels
  x_seq_validation_labels_exclude <- result$x_seq_labels_exclude
  result <- NULL
  
  x_seq_train <- keras::array_reshape(x_seq_train,dim=c(dim(x_seq_train)[1],
                                                        dim(x_seq_train)[2]*
                                                          dim(x_seq_train)[3]))
  x_seq_train_noevent <- keras::array_reshape(x_seq_train_noevent,dim=c(dim(x_seq_train_noevent)[1],
                                                                        dim(x_seq_train_noevent)[2]*
                                                                          dim(x_seq_train_noevent)[3]))
  x_seq_validation <- keras::array_reshape(x_seq_validation,dim=c(dim(x_seq_validation)[1],
                                                                  dim(x_seq_validation)[2]*
                                                                    dim(x_seq_validation)[3]))
  
  ## train isolation forest
  require(isotree)
  iso = isolation.forest(x_seq_train_noevent,
                         sample_size=sample_size,
                         ntrees = ntrees, 
                         ndim=ndim,
                         ntry=ntry,
                         max_depth=max_depth,
                         ncols_per_tree=ncols_per_tree,
                         nthreads=1
  )
	
  ## anomaly score for training data
  train_err <- predict(iso,x_seq_train)
  AUCtrain <- roc(x_seq_train_labels,train_err,direction="<")
	
  ## Find best threshold on training data
  thresh <- optimizeThreshold(x_seq_train_labels,train_err)
  
	## Apply to validation data  
  window <- 24*60
  validation_err <- NULL
  newdata <- x_seq_train_noevent
  for(jj in seq(from=1,by=window,to=dim(x_seq_validation)[1])){
    print(jj)
    sel <- jj:min((jj+window-1),dim(x_seq_validation)[1])
    validation_errjj <- predict(iso,x_seq_validation[sel,])
    validation_err <- c(validation_err,validation_errjj)
    outlying <- validation_errjj > thresh
    if(!all(outlying)){
      newdata <- rbind(newdata, x_seq_validation[sel,][!outlying,,drop=F])
      iso = isolation.forest(newdata,
                             sample_size=sample_size,
                             ntrees = ntrees, 
                             ndim=ndim,
                             ntry=ntry,
                             max_depth=max_depth,
                             ncols_per_tree=ncols_per_tree,
                             nthreads=1)
    }
  }
  AUCvalidation <- roc(x_seq_validation_labels,validation_err,direction="<")	
  
  plot(validation_err,type="l")
  points(which(x_seq_validation_labels),validation_err[x_seq_validation_labels],col="red")
  abline(h = thresh)
  
  gc()
  
  res <- calculateScore(x_seq_validation_labels,validation_err>thresh)
  return(-c(as.numeric(AUCvalidation$auc),res$SCORE,as.numeric(AUCtrain$auc)))
}
