## If not installed, do:
# install.packages("reticulate")
# install.packages("dplyr")
# install.packages("callr")
# install.packages("plotly")
# install.packages("ggsci")
# install.packages("emoa")
# install.packages("imputeTS")
# install.packages("keras")
# install.packages("abind")
# install.packages("pROC")
## Load the packages:
require(reticulate) 
require(dplyr) 
require(callr) 
require(keras)
require(plotly)
require(ggsci)
require(emoa)
library(imputeTS)
library(keras)
library(abind) 
library(pROC)

## Note, that this example requires python and the respective tensorflow / keras
## code to be set up correctly. 

## Note, that you have to make sure to use the right python
## version / environment (where the respective modules are installed)
## to execute the underlying run.py code,
## see e.g., 
#?reticulate::use_python

## Load a few R functions required for this example
source("getobjf.R")

source("misc.R")

source("plotting.R")

source("runAE.R")

source("runAE_test.R")

## configuration for the tuning problem
cfg <- list(
  ## Names of the parameters
  tunepars = c("batch_size","nfilter","lr",    "lrup",   "activation","drp",    "num_layers"),
  ## their lower bounds
  lower =    c(16,           10,       -5,      -2,        1,          -3,          1),
  ## their upper bounds
  upper =    c(1024,         100,      -2,       0,        3,          -0.3,        4), 
  ## their type
  type =    c("integer",    "integer","numeric","numeric","factor",   "numeric","integer"),
  ## transformations to apply
  transformations = 
    c(        trans_id,     trans_id, trans_10pow,trans_10pow,trans_id,trans_10pow,trans_id),
  ## another parameter that will not be tuned, but is fixed					 
  fixpars = list(),
  ## not used in this example (specify levels of categorical parameters):
  factorlevels = list(activation=c("relu","swish","sigmoid")),
  ## not used in this example (specify parameters that are relative to other parameters)
  relpars = list(),
  run=runAE
)

## penalty value if modeling fails
penalty <- -c(0.5,0,0.5)

## create an objective function that contains the parameterized learning process
objf <- get_objf(cfg,
                 600, # timeout in seconds
                 penalty) # and the penalty

## create an objective function that contains the parameterized learning process
cfg_test <- cfg
cfg_test$run <- runAE_test
objf_test <- get_objf(cfg_test,
                      1200, # timeout in seconds
                      penalty) # and the penalty

## the following are some optional checks if everything works as expected
## (evaluating twice at the lower and upper bound)
#res11 <- objf(matrix(cfg$lower,1),1)
#res12 <- objf(matrix(cfg$lower,1),2)
#res21 <- objf(matrix(cfg$upper,1),1)
#res22 <- objf(matrix(cfg$upper,1),2)


## install SPOT:
#install.packages("SPOT")
## load SPOT package
require(SPOT)

resultfile <- "resultFileAE.RData" 

if(!file.exists(resultfile)){
  keras_model_sequential() # only to initialize
  ## Run tuning with SPOT
  result <- spot(fun = objf,
                 lower=cfg$lower,
                 upper=cfg$upper,
                 control = list(types=cfg$type,
                                plots=T, # do progress plots
                                optimizer=optimDE, # optimize surrogate with DE
                                noise=TRUE, # problem is noisy
                                seedSPOT=1, # initial RNG seed for SPOT
                                seedFun=1, # initial, incremental seed for evaluation
                                designControl=list(size=50,replicates=1), # initial design
                                replicates=1, # evaluate each candidate 1 time
                                funEvals=150, # limit on number of evaluations
                                model=buildKriging, #Kriging / Gaussian process surogate
                                modelControl=list(target="ei",useLambda=T,reinterpolate=T), #Kriging configuration
                                optimizerControl=list(funEvals=400*length(cfg$lower)) # evaluation budget for DE
                 )
  )
  
  resultpp <- prepare_spot_result_plot(data=result,
                                       model=buildRandomForest,
                                       modelControl=list(types=cfg$type),
                                       namesx=cfg$tunepars,
                                       namesy="AUCvalidation",
                                       nameslog=c("F1validation","AUCtrain","duration")
  )
  resultpp2 <- prepare_spot_result_plot(data=result,
                                        model=buildKriging,
                                        modelControl=list(types=cfg$type,useLambda=T,reinterpolate=F),
                                        namesx=cfg$tunepars,
                                        namesy="AUCvalidation",
                                        nameslog=c("F1validation","AUCtrain","duration")
  )
  
  ## Test evaluation (unseen test data with resampled train+validation)
  resTestAE <- sapply(1:20,objf_test,x=resultpp$xbest)
  
  ## save result in RData file
  save(resultpp,resultpp2,result,resTestAE,file=resultfile)
}else{
  load(resultfile)
}

boxplot(t(resTestAE)[,1])
boxplot(t(resTestAE)[,2])
boxplot(t(resTestAE)[,3])
boxplot(t(resTestAE)[,4])
sum(resTestAE[4,]) #duration of test evaluation overall

## interactive plots
## surface plots via surrogate model
plot_surface(resultpp2,which = c(1,2))
plot_surface(resultpp2,which = c(1,3))
plot_surface(resultpp2,which = c(1,4))
plot_surface(resultpp2,which = c(1,5))
plot_surface(resultpp2,which = c(1,6))
plot_surface(resultpp2,which = c(1,7))

plot_surface(resultpp2,which = c(2,3))
plot_surface(resultpp2,which = c(2,4))
plot_surface(resultpp2,which = c(2,5))
plot_surface(resultpp2,which = c(2,6))
plot_surface(resultpp2,which = c(2,7))

plot_surface(resultpp2,which = c(3,4))
plot_surface(resultpp2,which = c(3,5))
plot_surface(resultpp2,which = c(3,6))
plot_surface(resultpp2,which = c(3,7))

plot_surface(resultpp2,which = c(4,5))
plot_surface(resultpp2,which = c(4,6))
plot_surface(resultpp2,which = c(4,7))

plot_surface(resultpp2,which = c(5,6))
plot_surface(resultpp2,which = c(5,7))

plot_surface(resultpp2,which = c(6,7))

## sensitivity
plot_sensitivity(resultpp2)

## parallel plot of the actual data
plot_parallel(resultpp2)
plot_parallel(resultpp2,yrange = c(-1,-0.95))

## Plot of pareto optimal data 
plot_pareto(resultpp2,yvar = 2)
plot_pareto(resultpp2,yvar = 3)
plot_pareto(resultpp2,yvar = 4)
